/* Copyright (c) 2018 Pomcor. Use subject to the PJCL License, https:pomcor.com/pjcl/pjcl-license.txt */

// StRnd = statistically (but not cryptographically) random

var pjclStRndBigInts1 = [];
var pjclStRndBigInts2 = []; 
var pjclStRndBigInts3 = []; 
function pjclStRnd(n, rndomBigInts) {
	for (var i = rndomBigInts.length; i < n; i++) {
		rndomBigInts[i] = pjclStRndLimb();
	}
	var z = [];
	for (i = 0; i < n; i++) {
		z[i] = rndomBigInts[i];
	}
	return z;
}

function pjclStRnd1(n) {
	return pjclStRnd(n, pjclStRndBigInts1);
}

function pjclStRnd2(n) {
	return pjclStRnd(n, pjclStRndBigInts2);
}

function pjclStRnd3(n) {
	return pjclStRnd(n, pjclStRndBigInts3);
}

function pjclBinaryOpTime(f, n, count) {
	var x = pjclStRnd1(n);
	var y = pjclStRnd2(n); 
	var t1 = performance.now();
	for (var i = 0; i < count; i++) {
		f(x,y);
	}
	var t2 = performance.now();
	var t = t2 - t1;
	return t;
}

function pjclBinaryOpTime(f, n, count) {
	var x = pjclStRnd1(n);
	var y = pjclStRnd2(n); 
	var t1 = performance.now();
	for (var i = 0; i < count; i++) {
		f(x,y);
	}
	var t2 = performance.now();
	var t = t2 - t1;
	return t;
}

function pjclUnaryOpTime(f, n, count) {
	var x = pjclStRnd1(n);
	var t1 = performance.now();
	for (var i = 0; i < count; i++) {
		f(x);
	}
	var t2 = performance.now();
	var t = t2 - t1;
	return t;
}

function pjclAverageOfTrials(z,trials) {
	var average;
	if (trials > 6) {
		var greatestval = smallestval = z[1];
		var greatestidx = smallestidx = 1;
		for (var j = 2; j < trials; j++) {
			if (z[j] > greatestval) {
				greatestval = z[j];
				greatestidx = j;
			}
			else if (z[j] < smallestval) {
				smallestval = z[j];
				smallestidx = j;
			}
		}
		if (greatestidx == 1 && smallestidx == 1) {
			average = greatestval;
		}
		else {
			var sum = 0;
			for (j = 1; j < trials; j++) {
				if (j != greatestidx && j != smallestidx) {
					sum = sum + z[j];
				}
			}
			average = sum/(trials - 3);	
		}
	}
	else if (trials > 4) {
		var sum = 0;
		for (var j = 1; j < trials; j++) {
			sum = sum + z[j];
		}
		average = sum/(trials - 1);
	}	
	else {
		var sum = 0;
		for (var j = 0; j < trials; j++) {
			sum = sum + z[j];
		}
		average = sum/trials;
	}
	return average;
}

function pjclBinaryOpTimeTrials(f, n, count, trials) {
	var x = pjclStRnd1(n);
	var y = pjclStRnd2(n);
	var z = [];
	for (var j = 0; j < trials; j++) { 
		var t1 = performance.now();
		for (var i = 0; i < count; i++) {
			f(x,y);
		}
		var t2 = performance.now();
		var t = t2 - t1;
		z[j] = t;
	}
	return pjclAverageOfTrials(z, trials);
}

function pjclUnaryOpTimeTrials(f, n, count, trials) {
	var x = pjclStRnd1(n);
	var z = [];
	for (var j = 0; j < trials; j++) { 
		var t1 = performance.now();
		for (var i = 0; i < count; i++) {
			f(x);
		}
		var t2 = performance.now();
		var t = t2 - t1;
		z[j] = t;
	}
	return pjclAverageOfTrials(z, trials);
}

function pjclKaratsubaMultIsWorthIt(n, trials) {
	var x = pjclStRnd1(n);
	var y = pjclStRnd2(n);
	var count = 1;
	var longMultTime, karatsubaTime;
	do {
		longMultTime = pjclBinaryOpTime(pjclMult_Long, n, count);
		pjclKaratsubaThresholdMult = n - 1;
		karatsubaTime = pjclBinaryOpTime(pjclMult_Karatsuba, n, count);
		if ((longMultTime + karatsubaTime) >= 200) {
			break;
		}
		count = count * 2;	
	}
	while (true);
	longMultTime = pjclBinaryOpTimeTrials(pjclMult_Long, n, count, trials);
	pjclKaratsubaThresholdMult = n - 1;
	karatsubaTime = pjclBinaryOpTimeTrials(pjclMult_Karatsuba, n, count, trials);
	return (karatsubaTime < longMultTime);
}

function pjclKaratsubaSqrIsWorthIt(n, trials) {
	var x = pjclStRnd1(n);
	var count = 1;
	var longMultTime, karatsubaTime;
	do {
		longMultTime = pjclUnaryOpTime(pjclSqr_Long, n, count);
		pjclKaratsubaThresholdSqr = n - 1;
		karatsubaTime = pjclBinaryOpTime(pjclSqr_Karatsuba, n, count);
		if ((longMultTime + karatsubaTime) >= 200) {
			break;
		}
		count = count * 2;	
	}
	while (true);
	longMultTime = pjclUnaryOpTimeTrials(pjclSqr_Long, n, count, trials);
	pjclKaratsubaThresholdSqr = n - 1;
	karatsubaTime = pjclUnaryOpTimeTrials(pjclSqr_Karatsuba, n, count, trials);
	return (karatsubaTime < longMultTime);
}

function pjclDefineKaratsubaThreshold(isWorthIt,trials) {
	var lowerbound = 4;
	var upperbound = 8;
        // upperbound - lowerbound will always be a power of 2
	console.log("Phase 1");
	while (!isWorthIt(upperbound, trials)) {
		console.log("upperbound "+ upperbound + " not worth it");
		lowerbound = lowerbound * 2;
		upperbound = upperbound * 2;
	}
	console.log("upperbound " + upperbound + " worth it");
	console.log("Phase 2");
	var middle;
	do {
		middle = (upperbound + lowerbound) / 2;
		if (isWorthIt(middle, trials)) {
			console.log("middle " + middle + " worth it");
			upperbound = middle;
		}
		else {
			console.log("middle " + middle + " not worth it");
			lowerbound = middle;
		}
		console.log("interval " + lowerbound + " " + upperbound);
	}
	while ((upperbound - lowerbound) > 2);
	middle = (upperbound + lowerbound) / 2;
	console.log("returning threshold " + middle);
	return lowerbound;
}

function pjclDefineKaratsubaMultThreshold() {
	return pjclDefineKaratsubaThreshold(pjclKaratsubaMultIsWorthIt,7);
}

function pjclDefineKaratsubaSqrThreshold() {
	return pjclDefineKaratsubaThreshold(pjclKaratsubaSqrIsWorthIt,7);
}
